/*
 * async.c - Utilities for async operations
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#include "async.h"
#include "signal.h"
#include "util.h"

#include <sys/wait.h>
#include <sys/signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>


AsyncOperation *async_operation_start(AsyncOperationPollFunc poll,
                                      AsyncOperationCleanupFunc cleanup,
                                      void *data1, void *data2,
                                      const char *path, char * const *args) {
    pid_t pid = fork();
    if (pid < 0) {
        warn("fork(2)");
        return NULL;
    } else if (pid == 0) {
        // child
        if (!GLOBAL_OPTS.verbose) {
            freopen("/dev/null", "w", stdout);
        }
        execvp(path, args);
        err(1, "execvp(2) for \"%s\"", path);
    } else {
        // parent
        AsyncOperation *op = malloc_checked(sizeof(AsyncOperation));
        op->pid = pid;
        op->data1 = data1;
        op->data2 = data2;
        op->poll_func = poll;
        op->cleanup_func = cleanup;
        op->done = false;
        return op;
    }
}

AsyncOperation *async_operation_take(AsyncOperationPollFunc poll,
                                     AsyncOperationCleanupFunc cleanup,
                                     void *data1, void *data2, pid_t pid) {
    AsyncOperation *op = malloc_checked(sizeof(AsyncOperation));
    op->pid = pid;
    op->data1 = data1;
    op->data2 = data2;
    op->poll_func = poll;
    op->cleanup_func = cleanup;
    op->done = false;
    return op;
}

AsyncOperationStatus async_operation_poll(AsyncOperation *op, bool hang) {
    if (op->done) {
        // polling a done process is a no-op
        return ASYNC_OPERATION_DONE;
    }
    pid_t proc = waitpid(op->pid, &op->status, hang ? 0 : WNOHANG);
    if (proc < 0) {
        warn("waitpid(2)");
        async_operation_cancel(op, false);
        return ASYNC_OPERATION_FAILED;
    } else if (proc > 0) {
        op->done = true;
    }
    if (op->poll_func && !op->poll_func(op)) {
        async_operation_cancel(op, false);
        return ASYNC_OPERATION_FAILED;
    }
    return op->done ? ASYNC_OPERATION_DONE : ASYNC_OPERATION_WORKING;
}

void async_operation_cancel(AsyncOperation *op, bool force) {
    if (op->done) {
        // cancelling a done process is a no-op
        return;
    }
    kill(op->pid, force ? SIGKILL : SIGTERM);
    if (!force) {
        int status;
        if (!timed_wait(op->pid, &status, 1000 /* 1ms */)) {
            warn("waitpid(2)");
        }
        if (!WIFEXITED(status) && !WIFSIGNALED(status)) {
            // force kill it if it refuses to exit itself
            async_operation_cancel(op, true);
        }
    }
    op->done = true;
}

void async_operation_free(AsyncOperation *op) {
    if (op) {
        if (!op->done) {
            async_operation_cancel(op, false);
        }
        if (op->cleanup_func) {
            op->cleanup_func(op);
        }
        free(op);
    }
}
