/*
 * drive.h - Simple abstractions for working with file systems and drives
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#ifndef INCLUDED_DRIVE_H
#define INCLUDED_DRIVE_H

#include "async.h"

#include <stdbool.h>
#include <sys/types.h>

typedef struct {
    char *name;
    char *path;
    char *fs;
    char *mnt;
} Drive;

/*
 * Attempt to list all connected USB drives and their file systems. The list is
 * sorted alphabetically.
 * Return: all connected USB (actually any serial drive, so SATA too) and their
 * file systems.
 */
Drive *drive_list_usb_drives(size_t *count);

/*
 * Free the array of drives DRIVES.
 */
void drive_free_drives(Drive *drives, size_t count);

/*
 * Use the fstyp(8) utility to try to find the FS for device file DEV. 
 * Return: The FS name allocated with malloc(3), or NULL if it could not be
 *         found.
 */
char *drive_guess_fs(const char *dev);

/*
 * Create an MSDOS (FAT32) FS on DRIVE.
 * Return: an operation on success, NULL otherwise
 */
AsyncOperation *drive_newfs_msdos(Drive *drive);

/*
 * Mount DRIVE onto a new temp. dir in /tmp.
 * Return: an operation on success, NULL otherwise
 */
AsyncOperation *drive_mount(Drive *drive);

/*
 * Unmount DRIVE. Force the operation if FORCE. Also rmdir(2) the mount point.
 * Return: an operation on success, NULL otherwise
 */
AsyncOperation *drive_unmount(Drive *drive, bool force);

#endif
