/*
 * screen.h - Simple menu system
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#ifndef INCLUDED_SCREEN_H
#define INCLUDED_SCREEN_H

#include "../button.h"
#include "../lcd.h"
#include "../util.h"

#include <time.h>
#include <libgpio.h>
#include <sqlite3.h>

typedef struct {
    sqlite3 *db;
    LCD *lcd;
    uint32_t temp;
    uint32_t humid;
    bool back_down;
    bool up_down;
    bool down_down;
    bool sel_down;
    bool force_draw;
} SensorState;

typedef struct _Screen Screen;
// should return true if we are to go back
typedef bool (*ScreenDispatchFunc)(Screen *, SensorState *);
typedef void (*ScreenCleanupFunc)(Screen *);

struct _Screen {
    char *name;
    ScreenDispatchFunc dispatch_func;
    ScreenCleanupFunc cleanup_func;
};

/*
 * Init Screen with NAME and the given callbacks.
 * Return: the new Screen
 */
void screen_init(Screen *screen, const char *name,
                 ScreenDispatchFunc dispatch_func,
                 ScreenCleanupFunc cleanup_func);

/*
 * Cleanup screen by calling its cleanup func and then freeing its name
 */
void screen_delete(Screen *screen);

typedef struct {
    sqlite3 *db;
    LCD *lcd;
    Button *back_btn;
    Button *up_btn;
    Button *down_btn;
    Button *sel_btn;
    Screen **screens;
    size_t screen_count;

    // internal
    ssize_t cur_screen;
    size_t cur_menu_line;
    size_t last_drawn_menu_line;
    bool force_draw;
} ScreenManager;

/*
 * Crate a new ScreenManager with the given DB, LCD and buttons. DB and LCD
 * still owned by the caller!
 * Return: the new ScreenManager
 */
ScreenManager *screen_manager_new(sqlite3 *db, LCD *lcd, gpio_handle_t handle,
                                  gpio_pin_t back_pin, gpio_pin_t up_pin,
                                  gpio_pin_t down_pin, gpio_pin_t sel_pin);

/*
 * Delete ScreenManager and the buttons it manages. This does NOT delete it's
 * LCD.
 */
void screen_manager_delete(ScreenManager *sm);

/*
 * Add SCREEN to SM. 
 */
void screen_manager_add(ScreenManager *sm, Screen *screen);

/*
 * Process input and, if needed, redraw the screen for ScreenManager. TEMP and
 * HUMID should be the last values read by the connected THS.
 */
void screen_manager_dispatch(ScreenManager *sm, uint32_t temp, uint32_t humid);

typedef struct {
    Screen parent;
    uint32_t last_temp;
    uint32_t last_humid;
    time_t last_min;
} StatsScreen;

/*
 * Create a new stats screen. This screen will display the current temp. and
 * humidity data from a THS sensor.
 */
StatsScreen *stats_screen_new(void);

#endif
